/* record audio
 * by Marek Zelem
 * (c)20.6.1999
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <sys/soundcard.h>

#define	REDRAW_PER_SECOND	4

#ifdef USE_CURSES
#include <ncurses/term.h>
#endif USE_CURSES

#include "wav.h"

int set_dsp( int fd, int rate, int channels, int bites )
{ int x,h;
	h=x=(bites==8?AFMT_U8:(bites==16?AFMT_S16_LE:-1));
	if( h<=0 || ioctl(fd,SNDCTL_DSP_SETFMT,&x)<0 || x!=h )
	{	printf("Can't set %d-bit sample size!\n",bites);
		return(-1);
	}
	x=channels-1;
	if( ioctl(fd,SNDCTL_DSP_STEREO,&x)<0 || x!=channels-1 )
	{	printf("Can't set %d channels!\n",channels);
		return(-1);
	}
	x=rate;
	if( ioctl(fd,SNDCTL_DSP_SPEED,&x)<0 || x!=rate )
	{	printf("Can't set %d Hz sample rate (%d)!\n",rate,x);
		return(-1);
	}
	return(0);
}

int reset_dsp( int fd )
{
	ioctl(fd,SNDCTL_DSP_SYNC);
	ioctl(fd,SNDCTL_DSP_RESET);
	return(0);
}

int get_dsp_bufsize( int fd )
{ int x;
	if( ioctl(fd,SNDCTL_DSP_GETBLKSIZE,&x)<0 )
	{	printf("Can't get block size!\n");
		return(-1);
	}
	return(x);
}

/* --------------------------------- */

int write_n_bytes( int f, char c, int len )
{ char b[1024];
	memset(b,c,1024);
	while( len>=1024 )
	{	write(f,b,1024);
		len-=1024;
	}
	if( len>0 )
		write(f,b,len);
	return(0);
}

int fd,fdo;
int rate=44100;
int bites=8;
int bytes;
int channels=1;
int use_chan=0;	/* vsetky */
int operation=0;	/* 0-nic, 1-plus -1-minus */
int quasi_stereo=0;
long quasi_delay=0;
int slash=1;
int bufsize;
long max_len=0;	/* no limit */
long out_len=0;
long pause_len;
unsigned char *buf=NULL;
unsigned char *obuf=NULL;
unsigned char *qdbuf=NULL;
char key[1];

//#define read_buf()	read(fd,buf,bufsize)
inline int read_buf( void )
{ int r,i,c;			/* !!! POZOR funguje to iba na 8 a 16 bit */
  signed short a,x;
	if( (r=read(fd,buf,bufsize))<=0 )	return(r);
	if( operation==1 )
	{	for(i=0;i<bufsize;)
		{	a=0;
			for(c=0;c<channels;c++)
			{	x=buf[i+c*bytes];
				if( bytes>1 )
					x+=(buf[i+c*bytes+1])<<8;
				a+=x;
			}
			a/=channels;
			buf[i+(use_chan-1)*bytes]=a&0x00ff;
			if( bytes>1 )
				buf[i+(use_chan-1)*bytes+1]=(a>>8)&0x00ff;
			i+=channels*bytes;
		}
	}
	else if( operation==-1 )
	{	for(i=0;i<bufsize;)
		{	a=0;
			for(c=0;c<channels;c++)
			{	if(c==use_chan-1)	continue;
				x=buf[i+c*bytes];
				if( bytes>1 )
					x+=(buf[i+c*bytes+1])<<8;
				a+=x;
			}
			a/=channels-1;
			if( bytes==1 )
				buf[i+(use_chan-1)*1]-=a-0x80;
			else
			{	x=*((signed short *)&(buf[i+(use_chan-1)*bytes]));
				x-=(signed short)(a&0x00ffff);
				*((signed short *)&(buf[i+(use_chan-1)*bytes]))
					=x;
			}
			i+=channels*bytes;
		}
	}
	return(r);
}

inline int write_buf( void )
{ int i,o,j;
  static int n_slash=0;
  static long prev_value=0;
  long value,m,delta;
  static long quasi_pos=0;
	if( slash==2 )
	{	for(i=0,o=0;i<bufsize;i++)
		{	if( n_slash==1 )	buf[i]=~buf[i];
			o++;
			if( o==channels*bytes )
			{	o=0; n_slash=1-n_slash;		}
		}
	}
	if( use_chan==0 )
	{	out_len+=bufsize;
		return(write(fdo,buf,bufsize));
	}
	if( quasi_stereo==1 )
	{	i=(use_chan-1)*bytes;
		for(o=0;i<bufsize;)
		{	memcpy(obuf+o,buf+i,bytes);
			o+=bytes;
			for(value=0,m=1,j=0;j<bytes;j++)
			{	value+=buf[i+j]*m; m<<=8;	}
			if( bytes==1 )	value-=0x80;
			delta=value-prev_value;
			delta*=3;
			if( bytes==1 )	delta+=0x80;
			for(j=0;j<bytes;j++)
			{	obuf[o+j]=delta & 0xff; delta>>=8;	}
			o+=bytes;
			prev_value=value;
			i+=channels*bytes;
		}
		out_len+=o;
		return(write(fdo,obuf,o));
	}
	if( quasi_stereo==2 )
	{	i=(use_chan-1)*bytes;
		for(o=0;i<bufsize;)
		{	memcpy(obuf+o,buf+i,bytes);
			o+=bytes;
			memcpy(obuf+o,qdbuf+quasi_pos,bytes);
			o+=bytes;
			memcpy(qdbuf+quasi_pos,buf+i,bytes);
			quasi_pos+=bytes;
			if( quasi_pos>=quasi_delay )	quasi_pos-=quasi_delay;
			i+=channels*bytes;
		}
		out_len+=o;
		return(write(fdo,obuf,o));
	}
	i=(use_chan-1)*bytes;
	for(o=0;i<bufsize;)
	{	memcpy(obuf+o,buf+i,bytes);
		o+=bytes;
		i+=channels*bytes;
	}
	out_len+=o;
	return(write(fdo,obuf,o));
}

int rec_record( void );
int rec_paused( void );

enum {	QUIT=0,
	PAUSE,
	RECORD,
     };
int state=PAUSE;
int rec_start( void )
{ char i_buf[bufsize];
  char o_buf[(quasi_stereo?2:1)*bufsize/channels];
  char d_buf[quasi_delay];
	if( quasi_delay>0 )
  	{	if( bytes==1 )	memset(d_buf,0x80,quasi_delay);
  		else		memset(d_buf,0,quasi_delay);
	}
	buf=i_buf;
	obuf=o_buf;
	qdbuf=d_buf;
	printf(	"Press space to start/pause recording\n"
		"      enter to insert pause\n"
		"      q     to quit program\n"
		"\n");
	while( state>QUIT )
	{	if( state==PAUSE )	state=rec_paused();
		if( state==RECORD )	state=rec_record();
	}
	obuf=buf=NULL;
	return(0);
}

int insert_pause( void )
{
	if( bytes==1 )	write_n_bytes(fdo,0x80,pause_len);
	else		write_n_bytes(fdo,0x00,pause_len);
	out_len+=pause_len;
	printf("Pause inserted\n");
	return(0);
}

int rec_record( void )
{
	printf("Recording...  \n");
	while(1)
	{	if( max_len>0 && out_len>=max_len )	return(QUIT);
		if( read_buf()<=0 )			return(QUIT);
		write_buf();
		if( read(0,key,1)==1 )
		{	if( *key==' ' )	return(PAUSE);
			if( *key=='q' )	return(QUIT);
			if( *key=='\r' ) insert_pause();
			if( *key==0 )	read(0,key,1);
		}
	}
}

int rec_paused( void )
{ unsigned char a[channels];
  int i,c,n;
  int prev_max[channels];
  static char *teplomer=
//  	"------------------------------------------------------------";
  	"--- Amplitude meter ------- middle --------- optimal ---MAX-";
  static char *bodky=
  	"............................................................";
#ifdef USE_CURSES
  int up_lines=0;
#endif USE_CURSES
while(1)
{
  	for(c=0;c<channels;c++)	a[c]=0x80;
  	for(n=0;n<rate/REDRAW_PER_SECOND;)
	{	if( read_buf()<=0 )			return(QUIT);
		for(c=0,i=bytes-1;i<bufsize;)
		{	if( bytes>1 )	buf[i]+=0x80;
			if( a[c]<buf[i] )	a[c]=buf[i];
			c++;
			if( c==channels )
			{	c=0;	n++;	}
			i+=bytes;
		}
		if( read(0,key,1)==1 )
		{	if( *key==' ' )	return(RECORD);
			if( *key=='q' )	return(QUIT);
			if( *key=='\r' )
			{	insert_pause();
#ifdef USE_CURSES
				up_lines++;
#endif USE_CURSES
			}
			if( *key==0 )	read(0,key,1);
		}
	}
#ifdef USE_CURSES
	while( up_lines>0 )
	{	putp(cursor_up); up_lines--;	}
	up_lines=0;
#endif USE_CURSES
	for(c=0;c<channels;c++)
	{	if( use_chan!=0 && use_chan!=c+1 )
			continue;
		i=(60*(a[c]-0x80))/128;
		if( i>0 ) i--;
		if( channels==1 )	printf("    ");
		else if( c<2 )		printf("  %c ","LR"[c]);
		else			printf(" %2d ",c+1);
		
		if( prev_max[c]>i && prev_max[c]<60 )
			printf("|%*.*s>%*.*s%*s|\n",i,i,teplomer,
				prev_max[c]-i,prev_max[c]-i,bodky,
				60-prev_max[c]-1,"");
//			printf("|%*.*s>%*s>%*s|\n",i,i,teplomer,
//				prev_max[c]-i-1,"",60-prev_max[c]-1,"");
		else
			printf("|%*.*s>%*s|\n",i,i,teplomer,60-i-1,"");
		prev_max[c]=i;
#if 0
		if( channels==1 )
		  printf("  %c |%*.*s>%*s|\n",' ',i,i,teplomer,60-i-1,"");
		else if( c<2 )
		  printf("  %c |%*.*s>%*s|\n","LR"[c],i,i,teplomer,60-i-1,"");
		else
		  printf(" %2d |%*.*s>%*s|\n",c+1,i,i,teplomer,60-i-1,"");
#endif
#ifdef USE_CURSES
		up_lines++;
#endif USE_CURSES
	}
//	fsync(1);
}
}

wavhead_t *wav_gen_head( unsigned long len )
{ static wavhead_t w={"RIFF",0,"WAVE","fmt ",16,1,0,0,0,1,0,"data",0};
	w.channels=channels;
	if( use_chan!=0 )	w.channels=1;
	if( quasi_stereo )	w.channels=2;
	w.sample_rate=rate;
	w.bites=bites;
	w.buffer_size=w.sample_rate * w.channels * w.bites/8 ;
	w.data_size=len;
	w.riff_size=w.data_size+sizeof(wavhead_t);
	return(&w);
}

int wav_write_head( void )
{ unsigned long len;
	if( (len=max_len)<=0 )	len=0xffffff00;
	return( write(fdo,wav_gen_head(len),sizeof(wavhead_t)) );
}

int wav_correct_head( void )
{ off_t off;
  int r;
	off=lseek(fdo,0,SEEK_CUR);
	lseek(fdo,0,SEEK_SET);
	r=write(fdo,wav_gen_head(out_len),sizeof(wavhead_t));
	lseek(fdo,off,SEEK_SET);
	return(r);
}

int wav_test_fmt( wavhead_t *w )
{
	if( strncmp(w->riff,"RIFF",4) )	return(0);
	if( strncmp(w->wave,"WAVE",4) )	return(0);
	if( strncmp(w->fmt ,"fmt ",4) )	return(0);
	if( strncmp(w->data,"data",4) )	return(0);
	if( w->format_type!=1 )		return(0);
	if( w->bites!=8 && w->bites!=16 )	return(0);
	return(1);
}

int wav_get_param( int f )
{ off_t off;
  wavhead_t w;
  	off=lseek(f,0,SEEK_CUR);
	if( read(f,&w,sizeof(wavhead_t))!=sizeof(wavhead_t) )
		goto nowave;
	if( !wav_test_fmt(&w) )
		goto nowave;
	channels=w.channels;
	rate=w.sample_rate;
	bites=w.bites;
	bytes=bites/8;
	max_len=w.data_size;
	if( use_chan>0 )	max_len/=channels;
	max_len*=(quasi_stereo?2:1);
	printf("Wave input file ( rate=%d , %d bit , %d channels )\n",
		rate,bites,channels);
	return(1);
nowave:
	lseek(f,off,SEEK_SET);
	return(0);
}

int main( int argc, char *argv[] )
{ long max_time=0;	/* pomocne */
  char *dsp_file="/dev/dsp";
  char *out_file=NULL;
  int pause_sec=3;
  int no_ioctl=0;
  int old_console_flags;
  struct termios oldterm,newterm;
  int a;

#ifdef USE_CURSES
	if( setupterm(NULL, 1, NULL)!=1 )
	{	printf("Ncurses error!\n");
//		return(0);
	}
#endif USE_CURSES
	printf("Record Audio by Marek Zelem (c)20.6.1999\n");
	a=1;
	while( a<argc )
	{	if( !strcmp(argv[a],"-M") )
		{	channels=1; use_chan=0;	}
		else if( !strcmp(argv[a],"-S") )
		{	channels=2; use_chan=0;	}
		else if( !strcmp(argv[a],"-L") )
		{	channels=2; use_chan=1;	}
		else if( !strcmp(argv[a],"-R") )
		{	channels=2; use_chan=2;	}
		else if( !strcmp(argv[a],"-Q") )
		{	if( use_chan==0 )	{ channels=1; use_chan=1; }
			quasi_stereo=1;
		}
		else if( a+1<argc && !strcmp(argv[a],"-D") )
		{	if( use_chan==0 )	{ channels=1; use_chan=1; }
			quasi_stereo=2;
			quasi_delay=atoi(argv[++a]);
		}
		else if( !strcmp(argv[a],"-ADD") )
		{	use_chan=1;	operation=1;	}
		else if( !strcmp(argv[a],"-SUB") )
		{	operation=-1;	}
		else if( a+1<argc && !strcmp(argv[a],"-s") )
		{	rate=atoi(argv[++a]);	}
		else if( a+1<argc && !strcmp(argv[a],"-b") )
		{	bites=atoi(argv[++a]);	}
		else if( a+1<argc && !strcmp(argv[a],"-sl") )
		{	slash=atoi(argv[++a]);	}
		else if( a+1<argc && !strcmp(argv[a],"-t") )
		{	max_time=atoi(argv[++a]);	}
		else if( a+1<argc && !strcmp(argv[a],"-l") )
		{	max_len=atoi(argv[++a]);	}
		else if( a+1<argc && !strcmp(argv[a],"-i") )
		{	dsp_file=argv[++a];	}
		else if( !strcmp(argv[a],"-noioctl") )
		{	no_ioctl=1;	}
		else if( a+1<argc && !strcmp(argv[a],"-o") )
		{	out_file=argv[++a];	}
		else if( a+1<argc && !strcmp(argv[a],"-p") )
		{	pause_sec=atoi(argv[++a]);	}
		else if( !strcmp(argv[a],"-start") )
		{	state=RECORD;	}
		else if( !strcmp(argv[a],"-h") )
		{	goto usage;	}
		else
		{	printf("Invalid argument '%s'\n",argv[a]);
		}
		a++;
	}
	if( out_file==NULL )
	{	printf("Output file not specified!\n");
		goto usage;
	}
	bytes=bites/8;
	if( bites!=bytes*8 )
	{	printf("Bites per sample must by n*8\n");
		return(0);
	}
	if( quasi_stereo && use_chan==0 )
	{	printf("Input must be only one channel "
			"if you use quasi stereo\n");
		return(0);
	}
	if( operation==-1 && use_chan==0 )
	{	printf("You must specifi channel\n");
		return(0);
	}
	if( (fd=open(dsp_file,O_RDONLY))<0 )
	{	printf("Can't open %s\n",dsp_file);
		return(-1);
	}
	if( !no_ioctl )
	{
		if( set_dsp(fd,rate,channels,bites)<0 )
		{	close(fd); return(0);	}
		if( (bufsize=get_dsp_bufsize(fd))<=0 )
		{	reset_dsp(fd);	close(fd);	return(0);	}
	}
	else
	{	wav_get_param(fd); bufsize=bytes*channels;	}
	if( (fdo=open(out_file,O_WRONLY|O_CREAT|O_TRUNC,0666))<0 )
	{	printf("Can't open %s\n",out_file);
		reset_dsp(fd); close(fd); return(-1);
	}
	max_time*=rate*channels*bites/8;
	if( max_len<=0 || (max_time>0 && max_time<max_len) )
		max_len=max_time;
	pause_len=pause_sec*rate*channels*bites/8;
	quasi_delay*= rate*1*bites/8000; /* zadane v ms */
	out_len=0;
	wav_write_head();

	tcgetattr(0,&oldterm);
	newterm=oldterm;
	newterm.c_lflag=0;
	newterm.c_iflag&=~(ICRNL|IGNCR|INLCR);
//	newterm.c_oflag=ONLCR;
	newterm.c_cc[VMIN]=1;
	newterm.c_cc[VTIME]=0;
	tcsetattr(0,TCSADRAIN,&newterm);
	old_console_flags=fcntl(0,F_GETFL,0);
	fcntl(0,F_SETFL,old_console_flags|O_NONBLOCK|O_NDELAY);
	rec_start();
	fcntl(0,F_SETFL,old_console_flags);
	tcsetattr(0,TCSADRAIN,&oldterm);

	wav_correct_head();

	reset_dsp(fd);
	close(fd);
	close(fdo);
	return(0);
usage:
	printf(	"usage: %s [options]\n",argv[0]);
	printf(	"options:\n"
		"   -M      - record Mono\n"
		"   -S      - record Stereo\n"
		"   -L      - record Left channel\n"
		"   -R      - record Right channel\n"
		"   -Q      - record Mono as Quasi stereo\n"
		"   -D msec - record Mono as Quasi stereo - delay msec\n"
		"   -ADD    - add all channels into one\n"
		"   -SUB    - aubtract other channels from selected\n"
		"   -s rate - set sample rate\n"
		"   -b 8/16 - set bites per sample\n"
		"   -sl num - slash number\n"
		"   -t time - limit sampling time\n"
		"   -l len  - limit output length\n"
		"   -i file - use file instead of /dev/dsp\n"
		"   -noioctl- don't ioctl on input file\n"
		"   -o file - output to file\n"
		"   -p sec  - pause length in seconds\n"
		"   -start  - start recording immediately\n"
		);
	return(0);
}

